# https://slurm.schedmd.com/squeue.html#lbAG
STATES_FINISHED = {
    "BOOT_FAIL",
    "CANCELLED",
    "COMPLETED",
    "DEADLINE",
    "FAILED",
    "NODE_FAIL",
    "OUT_OF_MEMORY",
    "PREEMPTED",
    "SPECIAL_EXIT",
    "TIMEOUT",
}
