"""Add Task.docs_info and Task.docs_link

Revision ID: 8f79bd162e35
Revises: a7f4d6137b53
Create Date: 2023-08-01 14:13:15.322398

"""
import sqlalchemy as sa
import sqlmodel
from alembic import op


# revision identifiers, used by Alembic.
revision = "8f79bd162e35"
down_revision = "a7f4d6137b53"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("task", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "docs_info", sqlmodel.sql.sqltypes.AutoString(), nullable=True
            )
        )
        batch_op.add_column(
            sa.Column(
                "docs_link", sqlmodel.sql.sqltypes.AutoString(), nullable=True
            )
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("task", schema=None) as batch_op:
        batch_op.drop_column("docs_link")
        batch_op.drop_column("docs_info")

    # ### end Alembic commands ###
