"""user settings project dir

Revision ID: 19eca0dd47a9
Revises: 8e8f227a3e36
Create Date: 2024-10-30 14:34:28.219355

"""
import sqlalchemy as sa
import sqlmodel
from alembic import op


# revision identifiers, used by Alembic.
revision = "19eca0dd47a9"
down_revision = "8e8f227a3e36"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("user_settings", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "project_dir",
                sqlmodel.sql.sqltypes.AutoString(),
                nullable=True,
            )
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("user_settings", schema=None) as batch_op:
        batch_op.drop_column("project_dir")

    # ### end Alembic commands ###
