"""Remove UserSettings.cache_dir

Revision ID: 316140ff7ee1
Revises: d256a7379ab8
Create Date: 2024-12-03 10:15:53.255958

"""
import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "316140ff7ee1"
down_revision = "d256a7379ab8"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("user_settings", schema=None) as batch_op:
        batch_op.drop_column("cache_dir")

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("user_settings", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "cache_dir", sa.VARCHAR(), autoincrement=False, nullable=True
            )
        )

    # ### end Alembic commands ###
